# coding=utf-8
"""Methods for parsing comfort-related data from EnergyPlus output files."""
from __future__ import division

import json

from ladybug.sql import SQLiteResult
from ladybug.datacollection import HourlyContinuousCollection
from ladybug.header import Header
from ladybug.datatype.speed import AirSpeed


def _parse_enclosure_info(enclosure_info, result_sql, epw, analysis_period=None,
                          default_air_speed=0.1, include_humidity=False,
                          use_10m_wind_speed=False):
    """Get lists of comfort-related data collections from an enclosure_info JSON.

    Args:
        enclosure_info: Path to a JSON file containing information about the radiant
            enclosure that sensor points belong to.
        result_sql: Path to an SQLite file that was generated by EnergyPlus.
            This file must contain hourly or sub-hourly results for zone comfort
            variables.
        epw: An EPW object that will be used to specify data for any sensor outside
            of any enclosure.
        analysis_period: An optional AnalysisPeriod to be applied to all results.
            If None, all data collections will be for the entire run period of
            the result_sql.
        default_air_speed: A single value or data collection to be used for all
            indoor air speed.
        include_humidity: Boolean to note whether data collections of humidity should
            be returned or not.
        use_10m_wind_speed: Boolean to note whether the meteorological wind speed
            should be used as-is for any outdoor sensors or whether it should be
            converted to ground-level speed (multiplying by 2/3).

    Returns:
        A tuple of 5 values.

        * pt_air_temps -- Data collections of air temperatures.

        * pt_rad_temps -- Data collections of long wave mean radiant temperature.

        * pt_humids - Data collections of relative humidity if include_humidity is True.

        * pt_speeds - Data collections of air speed values.

        * base_a_per - The AnalysisPeriod of the data in the result_sql.
    """
    # load all comfort-related outputs from the result_sql
    sql_obj = SQLiteResult(result_sql)
    air_temps = sql_obj.data_collections_by_output_name('Zone Mean Air Temperature')
    rad_temps = sql_obj.data_collections_by_output_name('Zone Mean Radiant Temperature')
    if include_humidity:
        humids = sql_obj.data_collections_by_output_name('Zone Air Relative Humidity')

    # check that EnergyPlus sql data is correct and note the analysis period
    assert len(air_temps) != 0, \
        'Input result-sql does not contain thermal comfort outputs.'
    assert isinstance(air_temps[0], HourlyContinuousCollection), 'EnergyPlus ' \
        'reporting frequency must be Hourly or Timestep to use thermal mapping. ' \
        'Not {}'.format(air_temps[0])
    base_a_per = air_temps[0].header.analysis_period

    # convert default air speed into a data collection if it's a list
    default_air_speed = _values_to_data(default_air_speed, base_a_per, AirSpeed, 'm/s')

    # parse the enclosure_info
    with open(enclosure_info) as json_file:
        enclosure_dict = json.load(json_file)

    # order the sql data based on the relevant zones from the enclosure_info
    rel_air_temps, rel_rad_temps, rel_humids, rel_speeds = [], [], [], []
    for zone_id in enclosure_dict['mapper']:
        zone_id = zone_id.upper()  # capitalize to match the output of EnergyPlus
        for data in air_temps:
            if data.header.metadata['Zone'] == zone_id:
                rel_air_temps.append(data)
                break
        for data in rad_temps:
            if data.header.metadata['Zone'] == zone_id:
                rel_rad_temps.append(data)
                break
        if include_humidity:
            for data in humids:
                if data.header.metadata['System'] == zone_id:
                    rel_humids.append(data)
                    break
        rel_speeds.append(default_air_speed)

    # if the enclosure info includes outdoor sensors, ensure epw data is added
    if enclosure_dict['has_outdoor']:
        _add_epw_data(epw, rel_air_temps, rel_rad_temps, rel_humids, rel_speeds,
                      base_a_per, use_10m_wind_speed)

    # apply the analysis periods if it is specified
    if analysis_period is not None and base_a_per != analysis_period:
        a_per = analysis_period
        rel_air_temps = [data.filter_by_analysis_period(a_per) for data in rel_air_temps]
        rel_rad_temps = [data.filter_by_analysis_period(a_per) for data in rel_rad_temps]
        if include_humidity:
            rel_humids = [data.filter_by_analysis_period(a_per) for data in rel_humids]
        new_rel_speeds = []
        for a_spd in rel_speeds:
            new_a_spd = a_spd.filter_by_analysis_period(a_per) \
                if isinstance(a_spd, HourlyContinuousCollection) else a_spd
            new_rel_speeds.append(new_a_spd)
        rel_speeds = new_rel_speeds

    # loop through the sensors and select the relevant data collections
    pt_air_temps, pt_rad_temps, pt_humids, pt_speeds = [], [], [], []
    for pt_i in enclosure_dict['sensor_indices']:
        pt_air_temps.append(rel_air_temps[pt_i])
        pt_rad_temps.append(rel_rad_temps[pt_i])
        if include_humidity:
            pt_humids.append(rel_humids[pt_i])
        pt_speeds.append(rel_speeds[pt_i])
    return pt_air_temps, pt_rad_temps, pt_humids, pt_speeds, base_a_per


def _values_to_data(values, base_period, data_type, data_units):
    """Load an array of values to a data collection.

    Args:
        values: An array of numbers.
        base_period: An AnalysisPeriod to be used for data collection.
        data_type: The class of the data type for the values.
        data_units: The units of the values.
    """
    if isinstance(values, list):
        header = Header(data_type(), data_units, base_period)
        return HourlyContinuousCollection(header, values)
    return values


def _add_epw_data(epw, rel_air_temps, rel_rad_temps, rel_humids, rel_speeds,
                  base_a_per, use_10m_wind_speed):
    """Add EPW data to zone data collections and align it with these collections."""
    rel_air_temps.append(epw.dry_bulb_temperature)
    out_l_mrt = (epw.dry_bulb_temperature + epw.sky_temperature) / 2
    rel_rad_temps.append(out_l_mrt)
    rel_humids.append(epw.relative_humidity)
    if use_10m_wind_speed:
        rel_speeds.append(epw.wind_speed)
    else:
        rel_speeds.append(epw.wind_speed * (2 / 3))  # conversion used by UTCI
    if not base_a_per.is_annual:  # apply sim analysis period to the annual EPW data
        rel_air_temps[-1] = rel_air_temps[-1].filter_by_analysis_period(base_a_per)
        rel_rad_temps[-1] = rel_rad_temps[-1].filter_by_analysis_period(base_a_per)
        rel_humids[-1] = rel_humids[-1].filter_by_analysis_period(base_a_per)
        rel_speeds[-1] = rel_speeds[-1].filter_by_analysis_period(base_a_per)
    if base_a_per.timestep != 1:  # interpolate the EPW data to timestep
        t_step = base_a_per.timestep
        rel_air_temps[-1] = rel_air_temps[-1].interpolate_to_timestep(t_step)
        rel_rad_temps[-1] = rel_rad_temps[-1].interpolate_to_timestep(t_step)
        rel_humids[-1] = rel_humids[-1].interpolate_to_timestep(t_step)
        rel_speeds[-1] = rel_speeds[-1].interpolate_to_timestep(t_step)



# coding=utf-8
"""Parameters for specifying acceptable thermal conditions using the PMV model."""
from __future__ import division
import re

from ._base import ComfortParameter
from ..pmv import ppd_threshold_from_comfort_class


class PMVParameter(ComfortParameter):
    """Parameters of PMV comfort.

    Args:
        ppd_comfort_thresh: A number between 5 and 100 that represents the upper
            threshold of PPD that is considered acceptable.
            Default is 10, which characterizes most buildings in the ASHRAE-55 and
            EN-15251 standards.
        humid_ratio_upper: A number between 0 and 1 indicating the upper limit of
            humidity ratio that is considered acceptable. Default is 1 for
            essentially no limit.
        humid_ratio_lower: A number between 0 and 1 indicating the lower limit of
            humidity ratio considered acceptable. Default is 0 for essentially
            no limit.
        still_air_threshold: The air speed threshold in m/s at which the standard
            effective temperature (SET) model will be used to correct for the
            cooling effect of elevated air speeds.
            Default is 0.1 m/s, which is the limit according to ASHRAE-55.

    Properties:
        * ppd_comfort_thresh
        * humid_ratio_upper
        * humid_ratio_low
        * still_air_threshold
    """
    _model = 'Predicted Mean Vote'
    __slots__ = ('_ppd_thresh', '_hr_upper', '_hr_lower', '_still_thresh')

    def __init__(self, ppd_comfort_thresh=None, humid_ratio_upper=None,
                 humid_ratio_lower=None, still_air_threshold=None):
        """Initialize PMV Parameters."""
        self.ppd_comfort_thresh = \
            ppd_comfort_thresh if ppd_comfort_thresh is not None else 10
        self._hr_upper = humid_ratio_upper if humid_ratio_upper is not None else 1
        self._hr_lower = humid_ratio_lower if humid_ratio_lower is not None else 0
        self._still_thresh = \
            still_air_threshold if still_air_threshold is not None else 0.1

        assert 0.008 <= self._hr_upper <= 1, \
            'humid_ratio_upper must be between 0.008 and 1. Got {}'.format(
                self._hr_upper)
        assert 0 <= self._hr_lower <= 0.005, \
            'humid_ratio_lower must be between 0 and 0.005. Got {}'.format(
                self._hr_lower)
        assert 0 <= self._still_thresh, \
            'still_air_threshold must be greater than 0. Got {}'.format(
                self._still_thresh)

    @classmethod
    def from_dict(cls, data):
        """Create a PMVParameter object from a dictionary.

        Args:
            data: A PMVParameter dictionary in following the format below.

        .. code-block:: python

            {
            'type': 'PMVParameter',
            'ppd_comfort_thresh': 20,
            'humid_ratio_upper': 0.12,
            'humid_ratio_lower': 0,
            'still_air_threshold': 0.2
            }
        """
        assert data['type'] == 'PMVParameter', \
            'Expected PMVParameter dictionary. Got {}.'.format(data['type'])
        ppd_comfort_thresh = data['ppd_comfort_thresh'] if \
            'ppd_comfort_thresh' in data else 10
        humid_ratio_upper = data['humid_ratio_upper'] if \
            'humid_ratio_upper' in data else 1
        humid_ratio_lower = data['humid_ratio_lower'] if \
            'humid_ratio_lower' in data else 0
        still_air_threshold = data['still_air_threshold'] if \
            'still_air_threshold' in data else 0.1
        return cls(ppd_comfort_thresh, humid_ratio_upper, humid_ratio_lower,
                   still_air_threshold)

    @classmethod
    def from_string(cls, pmv_parameter_string):
        """Create an PMVParameter object from an PMVParameter string."""
        str_pattern = re.compile(r"\-\-(\S*\s\S*)")
        matches = str_pattern.findall(pmv_parameter_string)
        par_dict = {item.split(' ')[0]: item.split(' ')[1] for item in matches}
        ppd_threshold = float(par_dict['ppd-threshold']) \
            if 'ppd-threshold' in par_dict else None
        hr_upper = float(par_dict['hr-upper']) \
            if 'hr-upper' in par_dict else None
        hr_lower = float(par_dict['hr-lower']) \
            if 'hr-lower' in par_dict else None
        still_air_threshold = float(par_dict['still-air-threshold']) \
            if 'still-air-threshold' in par_dict else None
        return cls(ppd_threshold, hr_upper, hr_lower, still_air_threshold)

    @property
    def ppd_comfort_thresh(self):
        """The threshold of the percentage of people dissatisfied (PPD) beyond which
        the conditions are not acceptable.  The default is 10%.
        """
        return self._ppd_thresh

    @ppd_comfort_thresh.setter
    def ppd_comfort_thresh(self, ppd):
        assert 5 <= ppd <= 100, \
            'ppd_comfort_thresh must be between 5 and 100. Got {}'.format(ppd)
        self._ppd_thresh = ppd

    @property
    def humid_ratio_upper(self):
        """A number representing the upper boundary of humidity ratio above which conditions
        are considered too humid to be comfortable.  The default is set to 0.03 kg
        wather/kg air.
        """
        return self._hr_upper

    @property
    def humid_ratio_lower(self):
        """A number representing the lower boundary of humidity ratio below which conditions
        are considered too dry to be comfortable.  The default is set to 0 kg wather/kg
        air.
        """
        return self._hr_lower

    @property
    def still_air_threshold(self):
        """A number representing the wind speed beyond which the formula for Standard
        Effective Temperature (SET) is used to determine PMV/PPD (as opposed to Fanger's
        original equation). The default is set to 0.1 m/s.
        """
        return self._still_thresh

    def set_ppd_comfort_thresh_from_comfort_class(self, comfort_class):
        """Set the PPD threshold given the EN-15251 comfort class."""
        self.ppd_comfort_thresh = ppd_threshold_from_comfort_class(comfort_class)

    def is_comfortable(self, ppd, humidity_ratio=0):
        """Determine if conditions are comfortable or not.

        Values are one of the following:

        * 0 = uncomfortable
        * 1 = comfortable
        """
        return 1 if (ppd <= self._ppd_thresh and humidity_ratio >= self._hr_lower and
                     humidity_ratio <= self._hr_upper) else 0

    def thermal_condition(self, pmv, ppd):
        """Determine whether conditions are cold, neutral or hot.

        Values are one of the following:

        * -1 = cold
        * 0 = netural
        * +1 = hot
        """
        if ppd >= self._ppd_thresh:
            return 1 if pmv > 0 else -1
        else:
            return 0

    def discomfort_reason(self, pmv, ppd, humidity_ratio=0):
        """Determine the reason why conditions are comfortable or not.

        Values are one of the following:

        * -2 = too dry
        * -1 = too cold
        * 0 = comfortable
        * +1 = too hot
        * +2 = too humid
        """
        if ppd >= self._ppd_thresh:
            return 1 if pmv > 0 else -1
        elif humidity_ratio < self._hr_lower:
            return -2
        elif humidity_ratio > self._hr_upper:
            return 2
        else:
            return 0

    def to_dict(self):
        """PMVParameter dictionary representation."""
        return {
            'type': 'PMVParameter',
            'ppd_comfort_thresh': self.ppd_comfort_thresh,
            'humid_ratio_upper': self.humid_ratio_upper,
            'humid_ratio_lower': self.humid_ratio_lower,
            'still_air_threshold': self.still_air_threshold
        }

    def __copy__(self):
        return PMVParameter(self.ppd_comfort_thresh, self.humid_ratio_upper,
                            self.humid_ratio_lower, self.still_air_threshold)

    def __repr__(self):
        """PMV comfort parameters representation."""
        return '--ppd-threshold {} --hr-upper {} ' \
            '--hr-lower {} --still-air-threshold {}'.format(
                self._ppd_thresh, self._hr_upper, self._hr_lower, self._still_thresh)
